/* Copyright (C) 2023 FIAS Frankfurt Institute for Advanced Studies, Frankfurt / Main
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Felix Weiglhofer [committer] */
#include "Options.h"

#include <boost/program_options.hpp>

#include <iostream>
#include <iterator>
#include <unordered_map>

using namespace cbm::algo;
namespace po = boost::program_options;

using fles::Subsystem;


namespace std
{
  template<class T>
  std::ostream& operator<<(std::ostream& os, const std::vector<T>& v)
  {
    copy(v.begin(), v.end(), std::ostream_iterator<T>(os, " "));
    return os;
  }
}  // namespace std


void validate(boost::any& v, const std::vector<std::string>& values, severity_level*, int)
{

  static const std::unordered_map<std::string, severity_level> levels {
    {"trace", severity_level::trace}, {"debug", severity_level::debug},     {"status", severity_level::status},
    {"info", severity_level::info},   {"warning", severity_level::warning}, {"error", severity_level::error},
    {"fatal", severity_level::fatal}};

  po::validators::check_first_occurrence(v);

  const std::string& s = po::validators::get_single_string(values);

  auto it = levels.find(s);

  if (it == levels.end()) throw po::validation_error(po::validation_error::invalid_option_value);

  v = it->second;
}

Options::Options(int argc, char** argv)
{

  po::options_description required("Required options");
  // clang-format off
  required.add_options()
    ("param-dir,p", po::value(&fParamsDir)->value_name("<folder>")->required(),
      "read program options from this folder")
    ("input-locator,i", po::value(&fInputLocator)->value_name("<locator>")->required(),
      "URI specifying input timeslice source")
  ;
  // clang-format on

  po::options_description generic("Other options");
  // clang-format off
  generic.add_options()
    ("output,o", po::value(&fOutputFile)->default_value("")->value_name("<file>"),
      "write results to file")
    ("device,d", po::value(&fDevice)->default_value("cpu")->value_name("<device>"),
      "select device (cpu, cuda0, cuda1, hip0, ...)")
    ("log-level,l", po::value(&fLogLevel)->default_value(info)->value_name("<level>"),
      "set log level (debug, info, warning, error, fatal)")
    ("monitor,m", po::value(&fMonitorUri)->value_name("<uri>")->implicit_value("file:cout"),
      "URI specifying monitor output (e.g. file:/tmp/monitor.txt, influx1:login:8086:cbmreco_status). Prints to cout when no argument is given. Monitor is disabled when flag is not set.")
    ("histogram", po::value(&fHistogramUri)->value_name("<uri>"), "URI to specify histogram server")
    ("log-file,L", po::value(&fLogFile)->value_name("<file>"),
      "write log messages to file")
    ("output-types,O", po::value(&fOutputTypes)->multitoken()->default_value({RecoData::Hit})->value_name("<types>"),
      "comma seperated list of reconstruction output types (hit, digi, ...)")
    ("steps", po::value(&fRecoSteps)->multitoken()->default_value({Step::Unpack, Step::LocalReco})->value_name("<steps>"),
      "comma seperated list of reconstruction steps (upack, digitrigger, localreco, ...)")
    ("systems,s", po::value(&fDetectors)->multitoken()->default_value({Subsystem::STS, Subsystem::TOF, Subsystem::BMON, Subsystem::MUCH, Subsystem::RICH, Subsystem::TRD, Subsystem::TRD2D})->value_name("<detectors>"),
      "comma seperated list of detectors to process (sts, mvd, ...)")
    ("child-id,c", po::value(&fChildId)->default_value("00")->value_name("<id>"), "online process id on node")
    ("num-ts,n", po::value(&fNumTimeslices)->default_value(-1)->value_name("<num>"),
      "Stop after <num> timeslices (-1 = all)")
    ("skip-ts", po::value(&fSkipTimeslices)->default_value(0)->value_name("<num>"),
      "Skip first <num> timeslices")
    ("omp", po::value(&fNumOMPThreads)->default_value(-1)->value_name("<num>"),
      "Set number of OpenMP threads (-1 = use OMP_NUM_THREADS environment variable)")
    ("times,t", po::bool_switch(&fCollectKernelTimes)->default_value(false),
      "print kernel times")
    ("dump-archive", po::bool_switch(&fDumpArchive)->default_value(false),
      "Dump archive content to stdout and exit. Provide archive with '-i'. (This is a hack to quick check archive content until we have proper tooling.)")
    ("help,h",
      "produce help message")
  ;
  // clang-format on

  po::options_description cmdline_options;
  cmdline_options.add(required).add(generic);

  po::variables_map vm;
  po::command_line_parser parser {argc, argv};
  parser.options(cmdline_options);
  try {
    auto result = parser.run();
    po::store(result, vm);
  }
  catch (const std::exception& e) {
    std::cerr << "Error: " << e.what() << std::endl;
    std::cerr << "Use '-h' to display all valid options." << std::endl;
    std::exit(EXIT_FAILURE);
  }

  if (vm.count("help") > 0) {
    std::cout << cmdline_options << std::endl;
    std::exit(EXIT_SUCCESS);
  }

  try {
    po::notify(vm);
  }
  catch (const po::required_option& e) {
    std::cerr << "Error: " << e.what() << std::endl;
    std::cerr << "Use '-h' to display all valid options." << std::endl;
    std::exit(EXIT_FAILURE);
  }
}
