/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CaTrackingMonitor.h
/// \brief  Monitor specialization for the tracking algorithm
/// \since  19.10.2023
/// \author S.Zharko <s.zharko@gsi.de>

#pragma once

#include "CaMonitor.h"

namespace cbm::algo::ca
{
  /// \enum  ECounter
  /// \brief Counter keys for the CA algo monitor
  enum class ECounter
  {
    TrackingCall,  ///< number of the routine calls
    SubTS,         ///< number of sub time-slices
    RecoTrack,     ///< number of reconstructed tracks
    RecoHit,       ///< number of reconstructed hits
    RecoHitUsed,   ///< number of used reconstructed hits
    Triplet,       ///< number of triplets
    // TODO: Provide counters vs. detector ID
    UndefinedMvdHit,   ///< number of undefined MVD hits
    UndefinedStsHit,   ///< number of undefined STS hits
    UndefinedMuchHit,  ///< number of undefined MuCh hits
    UndefinedTrdHit,   ///< number of undefined TRD hits
    UndefinedTofHit,   ///< number of undefined TOF hits
    kEND
  };

  /// \enum  ETimer
  /// \brief Timer keys for the CA algo monitor
  enum class ETimer
  {
    //Tracking,
    //PrepareInput,
    FindTracks,
    TrackFinder,
    TrackFitter,
    TripletConstruction,
    NeighboringTripletSearch,
    kEND
  };

  using TrackingMonitorData = MonitorData<ECounter, ETimer>;

  /// \class cbm::algo::ca::TrackingMonitor
  /// \brief A monitor specialization for cbm::algo::ca::Framework class
  class TrackingMonitor : public Monitor<ECounter, ETimer> {
  public:
    /// \brief Default constructor
    TrackingMonitor() : Monitor<ECounter, ETimer>("CA Framework Monitor")
    {
      SetCounterName(ECounter::TrackingCall, "full routine calls");
      SetCounterName(ECounter::RecoTrack, "reco tracks");
      SetCounterName(ECounter::RecoHit, "reco hits");
      SetCounterName(ECounter::Triplet, "triplets");
      SetCounterName(ECounter::RecoHitUsed, "used reco hits");
      SetCounterName(ECounter::SubTS, "sub-timeslices");
      SetCounterName(ECounter::UndefinedMvdHit, "undefined MVD hits");
      SetCounterName(ECounter::UndefinedStsHit, "undefined STS hits");
      SetCounterName(ECounter::UndefinedMuchHit, "undefined MuCh hits");
      SetCounterName(ECounter::UndefinedTrdHit, "undefined TRD hits");
      SetCounterName(ECounter::UndefinedTofHit, "undefined TOF hits");
      //SetTimerName(ETimer::Tracking, "full tracking");
      //SetTimerName(ETimer::PrepareInput, "input data initialization");
      SetTimerName(ETimer::FindTracks, "find tracks procedure");
      SetTimerName(ETimer::TrackFinder, "track finder");
      SetTimerName(ETimer::TrackFitter, "track fitter");
      SetTimerName(ETimer::TripletConstruction, "triplet construction");
      SetTimerName(ETimer::NeighboringTripletSearch, "neighboring triplet search");
      SetRatioKeys({ECounter::TrackingCall, ECounter::SubTS, ECounter::RecoTrack});
    }
  };
}  // namespace cbm::algo::ca
