/* Copyright (C) 2023 Facility for Antiproton and Ion Research in Europe, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Dominik Smith [committer] */

#include "Hitfind.h"

#include <chrono>

#include "log.hpp"

using namespace std;
using fles::Subsystem;

namespace cbm::algo::tof
{
  // -----   Execution   -------------------------------------------------------
  Hitfind::resultType Hitfind::operator()(gsl::span<CbmTofDigi> digiIn)
  {
    xpu::push_timer("TofHitfind");
    xpu::t_add_bytes(digiIn.size_bytes());

    // --- Output data
    resultType result = {};
    auto& clusterTs   = result.first;
    auto& monitor     = result.second;

    // Do calibration globally (optional, should not be used together with RPC-wise calibration)
    CalibRawDigis(digiIn, monitor);

    // Loop over the digis array and store the Digis in separate vectors for
    // each RPC modules
    for (int32_t idigi = 0; idigi < digiIn.size(); idigi++) {
      CbmTofDigi* pDigi = &(digiIn[idigi]);

      // These are doubles in the digi class
      const double SmType = pDigi->GetType();
      const double Sm     = pDigi->GetSm();
      const double Rpc    = pDigi->GetRpc();
      const int NbRpc     = fNbRpc[SmType];
      if (SmType >= fNbSm.size() || Sm * NbRpc + Rpc >= fNbSm[SmType] * NbRpc) {
        // Error already counted for monitoring during Digis calibration, so just discard here
        continue;
      }
      fStorDigi[SmType][Sm * NbRpc + Rpc].push_back(std::make_pair(pDigi, idigi));
    }

    std::vector<Hit> clustersFlat;
    std::vector<size_t> rpcSizes;   // nClusters per RPC
    std::vector<u32> rpcAddresses;  // RPC addresses

    // ---  RPC loop
    for (uint32_t SmType = 0; SmType < fNbSm.size(); SmType++) {
      const int NbRpc = fNbRpc[SmType];
      const int NbSm  = fNbSm[SmType];
      for (uint32_t Sm = 0; Sm < NbSm; Sm++) {
        for (uint32_t Rpc = 0; Rpc < NbRpc; Rpc++) {

          // Get digis
          std::vector<std::pair<CbmTofDigi*, int32_t>>& digiExp = fStorDigi[SmType][Sm * NbRpc + Rpc];

          // Build clusters
          std::vector<cbm::algo::tof::Hit> clusters = fAlgo[SmType][Sm * NbRpc + Rpc](digiExp);

          // Store hw address of partition
          // TODO: is this correct? Does this address make sense?
          u32 rpcAddress = CbmTofAddress::GetUniqueAddress(Sm, Rpc, 0, 0, SmType);
          rpcAddresses.push_back(rpcAddress);

          // store partition size
          rpcSizes.push_back(clusters.size());

          // Append clusters to output
          clustersFlat.insert(clustersFlat.end(), clusters.begin(), clusters.end());

          // Clear digi storage
          digiExp.clear();
        }
      }
    }

    // Monitoring
    monitor.fTime     = xpu::pop_timer();
    monitor.fNumDigis = digiIn.size();
    monitor.fNumHits  = clustersFlat.size();

    // Create ouput vector
    clusterTs = PartitionedVector(std::move(clustersFlat), rpcSizes, rpcAddresses);

    return result;
  }
  // ----------------------------------------------------------------------------


  // -----   Initialisation   ---------------------------------------------------
  void Hitfind::Init()
  {
    //Number of SMs per super module type
    fNbSm = fTofConfig.NbSm;

    //Number of RPCs per super module type
    fNbRpc = fTofConfig.NbRpc;

    //Prepare digi storage
    fStorDigi.resize(fNbSm.size());

    // Create one algorithm per RPC for TOF and configure it with parametersa
    for (int32_t SmType = 0; SmType < fNbSm.size(); SmType++) {

      int32_t NbSm  = fNbSm[SmType];
      int32_t NbRpc = fNbRpc[SmType];
      fStorDigi[SmType].resize(NbSm * NbRpc);

      for (int32_t Sm = 0; Sm < NbSm; Sm++) {
        for (int32_t Rpc = 0; Rpc < NbRpc; Rpc++) {

          std::unique_ptr<cbm::algo::tof::ClusterizerRpcPar> par(new cbm::algo::tof::ClusterizerRpcPar());

          HitfindSetup::Rpc rpcPar = fTofConfig.rpcs[SmType][Sm * NbRpc + Rpc];
          par->fDeadStrips         = rpcPar.deadStrips;
          par->fPosYMaxScal        = rpcPar.posYMaxScal;
          par->fdMaxTimeDist       = rpcPar.maxTimeDist;
          par->fdMaxSpaceDist      = rpcPar.maxSpaceDist;
          par->fSigVel             = rpcPar.sigVel;
          par->fCPTOffYBinWidth    = rpcPar.CPTOffYBinWidth;
          par->fCPTOffY            = rpcPar.CPTOffY;
          par->fCPTOffYRange       = rpcPar.CPTOffYRange;
          par->fTimeRes            = rpcPar.timeRes;
          par->numClWalkBinX       = rpcPar.numClWalkBinX;
          par->TOTMax              = rpcPar.TOTMax;
          par->TOTMin              = rpcPar.TOTMin;
          par->swapChannelSides    = rpcPar.swapChannelSides;
          par->channelDeadtime     = rpcPar.channelDeadtime;

          int32_t NbChan = rpcPar.chanPar.size();
          par->fChanPar.resize(NbChan);

          for (int32_t Ch = 0; Ch < NbChan; Ch++) {

            HitfindSetup::Channel chanPar   = rpcPar.chanPar[Ch];
            const double* tra_ptr           = chanPar.cell.translation.data();
            const double* rot_ptr           = chanPar.cell.rotation.data();
            par->fChanPar[Ch].address       = chanPar.address;
            par->fChanPar[Ch].cell.pos      = ROOT::Math::XYZVector(tra_ptr[0], tra_ptr[1], tra_ptr[2]);
            par->fChanPar[Ch].cell.rotation = ROOT::Math::Rotation3D(&rot_ptr[0], &rot_ptr[9]);
            par->fChanPar[Ch].cell.sizeX    = chanPar.cell.sizeX;
            par->fChanPar[Ch].cell.sizeY    = chanPar.cell.sizeY;
            par->fChanPar[Ch].fvCPTOff      = chanPar.vCPTOff;
            par->fChanPar[Ch].fvCPTotGain   = chanPar.vCPTotGain;
            par->fChanPar[Ch].fvCPTotOff    = chanPar.vCPTotOff;
            par->fChanPar[Ch].fvCPWalk      = chanPar.vCPWalk;
          }
          fAlgo[SmType][Sm * NbRpc + Rpc].SetParams(std::move(par));
        }
      }
    }
    L_(info) << "--- Configured hitfinder algorithms for TOF.";
    L_(info) << "==================================================";
  }
  // ----------------------------------------------------------------------------


  // -------------------------- Calibration -------------------------------------
  void Hitfind::CalibRawDigis(gsl::span<CbmTofDigi> digiVec, HitfindMonitorData& monitor)
  {
    // channel deadtime map
    std::map<int32_t, double> mChannelDeadTime;

    for (size_t iDigi = 0; iDigi < digiVec.size(); iDigi++) {

      CbmTofDigi pDigi    = digiVec[iDigi];
      const double SmType = pDigi.GetType();
      const double Sm     = pDigi.GetSm();
      const double Rpc    = pDigi.GetRpc();
      const double Chan   = pDigi.GetChannel();
      const double Side   = pDigi.GetSide();
      const int NbRpc     = fNbRpc[SmType];

      auto& rpcs = fTofConfig.rpcs;
      if (SmType >= rpcs.size() || Sm * NbRpc + Rpc >= rpcs.at(SmType).size()) {
        monitor.fDigiCalibUnknownRPC++;
        continue;
      }

      HitfindSetup::Rpc rpcPar      = fTofConfig.rpcs.at(SmType).at(Sm * NbRpc + Rpc);
      HitfindSetup::Channel chanPar = rpcPar.chanPar[Chan];

      if (rpcPar.swapChannelSides && 5 != SmType && 8 != SmType) {
        pDigi.SetAddress(Sm, Rpc, Chan, (0 == Side) ? 1 : 0, SmType);
      }

      // Check dead time
      const int32_t iAddr = pDigi.GetAddress();
      auto it             = mChannelDeadTime.find(iAddr);
      if (it != mChannelDeadTime.end() && pDigi.GetTime() <= it->second) {
        it->second = pDigi.GetTime() + rpcPar.channelDeadtime;
        continue;
      }
      mChannelDeadTime[iAddr] = pDigi.GetTime() + rpcPar.channelDeadtime;

      // Create calibrated digi
      CbmTofDigi pCalDigi(pDigi);

      // calibrate Digi Time
      pCalDigi.SetTime(pCalDigi.GetTime() - chanPar.vCPTOff[Side]);

      // subtract Offset
      double dTot = pCalDigi.GetTot() - chanPar.vCPTotOff[Side];
      if (dTot < 0.001) dTot = 0.001;
      // calibrate Digi ToT
      pCalDigi.SetTot(dTot * chanPar.vCPTotGain[Side]);

      // walk correction
      std::vector<double>& walk = chanPar.vCPWalk[Side];
      const double dTotBinSize  = (rpcPar.TOTMax - rpcPar.TOTMin) / rpcPar.numClWalkBinX;
      int32_t iWx               = (int32_t)((pCalDigi.GetTot() - rpcPar.TOTMin) / dTotBinSize);

      if (0 > iWx) { iWx = 0; }
      if (iWx >= rpcPar.numClWalkBinX) { iWx = rpcPar.numClWalkBinX - 1; }

      const double dDTot = (pCalDigi.GetTot() - rpcPar.TOTMin) / dTotBinSize - (double) iWx - 0.5;
      double dWT         = walk[iWx];

      // linear interpolation to next bin
      if (dDTot > 0) {
        if (iWx < rpcPar.numClWalkBinX - 1) { dWT += dDTot * (walk[iWx + 1] - walk[iWx]); }
      }
      else {
        if (0 < iWx) { dWT -= dDTot * (walk[iWx - 1] - walk[iWx]); }
      }
      pCalDigi.SetTime(pCalDigi.GetTime() - dWT);  // calibrate Digi Time
      digiVec[iDigi] = pCalDigi;
    }

    /// Sort the buffers of hits due to the time offsets applied
    std::sort(digiVec.begin(), digiVec.end(),
              [](const CbmTofDigi& a, const CbmTofDigi& b) -> bool { return a.GetTime() < b.GetTime(); });
  }

}  // namespace cbm::algo::tof
