/* Copyright (C) 2023 Facility for Antiproton and Ion Research in Europe, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Dominik Smith [committer] */
#ifndef CBM_ALGO_DETECTOR_TOF_HITFIND_SETUP_H
#define CBM_ALGO_DETECTOR_TOF_HITFIND_SETUP_H

#include <array>
#include <map>
#include <string>
#include <vector>

#include "Definitions.h"
#include "config/Property.h"

namespace cbm::algo::tof
{

  /**
   * @brief Hitfind setup / Hardware cabling for TOF
   * Used to create the hardware mapping for the TOF hitfinder.
   */
  struct HitfindSetup {

    struct Cell {
      double sizeX;
      double sizeY;
      std::array<double, 3> translation;
      std::array<double, 9> rotation;

      static constexpr auto Properties =
        std::make_tuple(config::Property(&Cell::sizeX, "sizeX", "size in X direction"),
                        config::Property(&Cell::sizeY, "sizeY", "size in Y direction"),
                        config::Property(&Cell::translation, "translation", "Translation vector", YAML::Flow),
                        config::Property(&Cell::rotation, "rotation", "Rotation matrix", YAML::Flow));
    };

    struct Channel {
      std::vector<double> vCPTOff;
      std::vector<double> vCPTotGain;
      std::vector<double> vCPTotOff;
      std::vector<std::vector<double>> vCPWalk;
      i32 address;
      Cell cell;

      static constexpr auto Properties =
        std::make_tuple(config::Property(&Channel::vCPTOff, "vCPTOff", "CPT offset"),
                        config::Property(&Channel::vCPTotGain, "vCPTotGain", "CP time over threshold gain"),
                        config::Property(&Channel::vCPTotOff, "vCPTotOff", "CP time over threshold offset"),
                        config::Property(&Channel::vCPWalk, "vCPWalk", "CP walk correction", YAML::Flow),
                        config::Property(&Channel::address, "address", "unique address"),
                        config::Property(&Channel::cell, "cell", "cell parameters"));
    };

    struct Rpc {
      u32 deadStrips;
      double posYMaxScal;
      double maxTimeDist;
      double maxSpaceDist;
      double sigVel;
      double timeRes;
      i32 numClWalkBinX;
      double TOTMax;
      double TOTMin;
      bool swapChannelSides;
      double channelDeadtime;
      double CPTOffYBinWidth;
      double CPTOffYRange;

      std::vector<double> CPTOffY;
      std::vector<Channel> chanPar;

      static constexpr auto Properties =
        std::make_tuple(config::Property(&Rpc::deadStrips, "deadStrips", "bit mask for dead strips"),
                        config::Property(&Rpc::posYMaxScal, "posYMaxScal", "maximum value of y position"),
                        config::Property(&Rpc::maxTimeDist, "maxTimeDist", "maximum time distance"),
                        config::Property(&Rpc::maxSpaceDist, "maxSpaceDist", "maximum space distance"),
                        config::Property(&Rpc::sigVel, "sigVel", "signal velocity"),
                        config::Property(&Rpc::timeRes, "timeRes", "time resolution"),
                        config::Property(&Rpc::numClWalkBinX, "numClWalkBinX", "number of walk correction bins"),
                        config::Property(&Rpc::TOTMax, "TOTMax", "maximum time over threshold"),
                        config::Property(&Rpc::TOTMin, "TOTMin", "minimum time over threshold"),
                        config::Property(&Rpc::swapChannelSides, "swapChannelSides", "flag for swapping channel sides"),
                        config::Property(&Rpc::channelDeadtime, "channelDeadtime", "channel dead time"),
                        config::Property(&Rpc::CPTOffYBinWidth, "CPTOffYBinWidth", "CPT Y offset bin width"),
                        config::Property(&Rpc::CPTOffYRange, "CPTOffYRange", "CPT Y offset range"),
                        config::Property(&Rpc::CPTOffY, "CPTOffY", "CPT Y offset array"),
                        config::Property(&Rpc::chanPar, "chanPar", "channel parameters"));
    };

    std::vector<std::vector<Rpc>> rpcs;
    std::vector<int32_t> NbSm;
    std::vector<int32_t> NbRpc;

    static constexpr auto Properties = std::make_tuple(
      config::Property(&HitfindSetup::rpcs, "rpcs", "Parameters of RPCs"),
      config::Property(&HitfindSetup::NbSm, "NbSm", "Number of SMs per super module type", {}, YAML::Flow),
      config::Property(&HitfindSetup::NbRpc, "NbRpc", "Number of RPCs per super module type", {}, YAML::Flow));
  };


}  // namespace cbm::algo::tof

#endif  // CBM_ALGO_DETECTOR_TOF_HITFIND_SETUP_H
