/* Copyright (C) 2022-2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// @file   CbmL1DetectorID.h TODO: Rename to CbmCaDefs.h !
/// @brief  Implementation of L1DetectorID enum class for CBM
/// @author S.Zharko
/// @since  01.12.2022

#ifndef CbmL1DetectorID_h
#define CbmL1DetectorID_h 1

#include <string>

#include "CaConstants.h"
#include "CaEnumArray.h"

namespace cbm::ca
{
  namespace phys      = cbm::algo::ca::constants::phys;
  namespace constants = cbm::algo::ca::constants;
  namespace clrs      = cbm::algo::ca::constants::clrs;
  namespace ca        = cbm::algo::ca;  ///< CA core namespace (TO BE USED IN CBM-specific interfaces)
}  // namespace cbm::ca

namespace cbm::algo::ca
{
  /// \enum  cbm::algo::ca::EDetectorID
  /// \brief Enumeration for the detector subsystems used in L1 tracking
  /// It is important for the subsystems to be specified in the actual order. The order is used
  /// for the ca::Station array filling.
  /// \note The enumeration must not contain jumps in the ordering and the first entry must be equal 0
  enum class EDetectorID
  {
    kMvd = 0,
    kSts,
    kMuch,
    kTrd,
    kTof,
    kEND  ///< End of enumeration
  };
}  // namespace cbm::algo::ca

/// *************************************************
/// **    Detector-dependent common definitions    **
/// *************************************************

class CbmMvdPoint;
class CbmStsPoint;
class CbmMuchPoint;
class CbmTrdPoint;
class CbmTofPoint;

class CbmMvdHit;
class CbmStsHit;
class CbmMuchPixelHit;
class CbmTrdHit;
class CbmTofHit;

namespace cbm::ca
{
  /// @brief  Alias to array, indexed by L1DetectorID enum
  /// @note   To be used only in CBM-specific code
  template<typename T>
  using DetIdArr_t = ca::EnumArray<ca::EDetectorID, T>;

  /// @brief  List of

  /// @struct DetIdTypeArr_t
  /// @brief  Array of types, indexed by L1DetectorID enum
  ///
  /// The array of types allows to treat different types of detector in a uniform manner
  /// Example:
  ///   using HitTypes_t = DetIdTypeArr_t<CbmMvdHit, CbmStsHit, CbmMuchPixelHit, CbmTrdHit, CbmTofHit>;
  ///   ...
  ///   HitTypes_t::at<L1DetectorID::kSts> hit;  // Sts hit
  template<class... Types>
  struct DetIdTypeArr_t {
    template<ca::EDetectorID DetID>
    using at                          = std::tuple_element_t<static_cast<std::size_t>(DetID), std::tuple<Types...>>;
    static constexpr std::size_t size = sizeof...(Types);
  };

  /// @brief Names of detector subsystems
  /// @note  These names are used for data branches IO, thus any modification can lead to the
  ///        read-out corruption.
  constexpr DetIdArr_t<const char*> kDetName = {{"MVD", "STS", "MUCH", "TRD", "TOF"}};

  /// @brief Name of hit branches for each detector
  constexpr DetIdArr_t<const char*> kDetHitBrName = {{"MvdHit", "StsHit", "MuchPixelHit", "TrdHit", "TofHit"}};

  /// @brief Name of point branches for each detector
  constexpr DetIdArr_t<const char*> kDetPointBrName = {{"MvdPoint", "StsPoint", "MuchPoint", "TrdPoint", "TofPoint"}};

  /// @brief Name

  /// @brief Types of MC point objects for each detector
  using PointTypes_t = DetIdTypeArr_t<CbmMvdPoint, CbmStsPoint, CbmMuchPoint, CbmTrdPoint, CbmTofPoint>;

  /// @brief Types of hit objects for each detector
  using HitTypes_t = DetIdTypeArr_t<CbmMvdHit, CbmStsHit, CbmMuchPixelHit, CbmTrdHit, CbmTofHit>;

}  // namespace cbm::ca

/// @brief Enumeration for different tracking running modes
enum class ECbmCaTrackingMode
{
  kSTS,  ///< Local tracking in CBM (STS + MVD), results stored to the StsTrack branch
  kMCBM  ///< Global tracking in mCBM (STS, MuCh, TRD, TOF), results stored to GlobalTrack branch
};


#endif  // CbmL1DetectorID_h
