/* Copyright (C) 2023 Facility for Antiproton and Ion Research in Europe, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Pierre-Alain Loizeau [committer] */

#ifndef CBM_SERVICES_HISTSERV_APP_APPLICATION_H
#define CBM_SERVICES_HISTSERV_APP_APPLICATION_H 1

#include "THttpServer.h"
#include "TObjArray.h"

#include <string>
#include <thread>
#include <zmq.hpp>

#include "ProgramOptions.h"

class TNamed;
class TCanvas;

namespace cbm::services::histserv
{

  class Application {


  public:
    /** @brief Standard constructor, initialises the application
     ** @param opt  **/
    explicit Application(ProgramOptions const& opt);

    /** @brief Copy constructor forbidden **/
    Application(const Application&) = delete;

    /** @brief Assignment operator forbidden **/
    void operator=(const Application&) = delete;

    /** @brief Destructor **/
    ~Application();

    /** @brief Run the application **/
    void Exec();

    void UpdateHttpServer();

  private:
    //const std::string& ConfigFile() const;

    bool ReceiveData(zmq::message_t& msg);
    bool ReceiveHistoConfig(zmq::message_t& msg);
    bool ReceiveCanvasConfig(zmq::message_t& msg);
    bool ReceiveConfigAndData(std::vector<zmq::message_t>& vMsg);

    template<class HistoT>
    bool ReadHistogram(HistoT* pHist);
    int FindHistogram(const std::string& name);
    bool PrepareCanvas(uint32_t uCanvIdx);

    bool ResetHistograms();
    bool SaveHistograms();

  private:
    ProgramOptions const& fOpt;      ///< Program options object
    THttpServer* fServer = nullptr;  ///< ROOT Histogram server (JSroot)
    std::thread fThread;
    bool fStopThread = false;

    /// Interface
    zmq::context_t fZmqContext {1};
    zmq::socket_t fZmqSocket {fZmqContext, ZMQ_PULL};

    /// Array of histograms with unique names
    TObjArray fArrayHisto;
    /// Vector of string with ( HistoName, FolderPath ) to configure the histogram
    std::vector<std::pair<std::string, std::string>> fvpsHistosFolder = {};
    /// Vector of string pairs with ( CanvasName, CanvasConfig ) to configure the canvases and histos within
    /// Format of Can config is "Name;Title;NbPadX(U);NbPadY(U);ConfigPad1(s);....;ConfigPadXY(s)"
    /// Format of Pad config is "GrixX(b),GridY(b),LogX(b),LogY(b),LogZ(b),HistoName(s),DrawOptions(s)"
    std::vector<std::pair<std::string, std::string>> fvpsCanvasConfig = {};
    std::vector<bool> fvbCanvasReady                                  = {};
    bool fbAllCanvasReady                                             = false;

    std::vector<std::pair<TNamed*, std::string>> fvHistos  = {};  ///< Vector of Histos pointers and folder path
    std::vector<bool> fvbHistoRegistered                   = {};
    bool fbAllHistosRegistered                             = false;
    std::vector<std::pair<TCanvas*, std::string>> fvCanvas = {};  ///< Vector of Canvas pointers and folder path
    std::vector<bool> fvbCanvasRegistered                  = {};
    bool fbAllCanvasRegistered                             = false;

    /// Internal status
    uint32_t fNMessages = 0;
  };

}  // namespace cbm::services::histserv

#endif /* CBM_SERVICES_HISTSERV_APP_APPLICATION_H */
